/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer: Benoit Pelletier
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.ant.cluster;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;


import org.objectweb.jonas.ant.jonasbase.JEcho;
import org.objectweb.jonas.ant.jonasbase.JMkdir;
import org.objectweb.jonas.ant.jonasbase.JTouch;
import org.objectweb.jonas.ant.jonasbase.Tasks;

/**
 * Allow to launch the cluster nodes
 * @author Benoit Pelletier
 */
public class Script extends Tasks {

    /**
     * Info for the logger
     */
    private static final String INFO = "[Script] ";

    /**
     * Name of the unix script file
     */
    private static final String UNIX_SCRIPT_FILE = "runCluster.sh";

    /**
     * Name of the unix script file
     */
    private static final String WIN_SCRIPT_FILE = "runCluster.cmd";

    /**
     * Script file Unix
     */
    private File unixScriptFile = null;

    /**
     * Script file
     */
    private File winScriptFile = null;


    /**
     * Domain name
     */
    private String domainName = null;

    /**
     * Workers list
     */
    private List nodesList = null;


    /**
     * Default constructor
     */
    public Script() {
        super();
        nodesList = new ArrayList();
    }

    /**
     * Creation of the mod_jk files
     * @param destDir destination directory
     */
    public void createFiles(String destDir) {
        JMkdir mkdir = new JMkdir();
        mkdir.setDestDir(new File(destDir));
        addTask(mkdir);

        JTouch touchUnix = new JTouch();
        unixScriptFile = new File(destDir + "/" + UNIX_SCRIPT_FILE);
        touchUnix.setDestDir(unixScriptFile);
        addTask(touchUnix);

        JTouch touchWin = new JTouch();
        winScriptFile = new File(destDir + "/" + WIN_SCRIPT_FILE);
        touchWin.setDestDir(winScriptFile);
        addTask(touchWin);

   }

    /**
     * Add a node
     * @param prefixName prefix
     * @param jonasBase JONAS_BASE dir
     * @return node name
     */
    public String addNode(String prefixName, String jonasBase) {
        Node node = new Node();
        int index = nodesList.size() + 1;
        node.setName(prefixName + index);
        node.setJonasBase(jonasBase);
        nodesList.add(node);
        return node.getName();
    }

    /**
     * get a node launching script
     * @param node node
     * @return definition of the node
     */
    private String getLinuxNodeDef(Node node) {


        String nodeDef = "\n"
            +
            "# -----------------------" + "\n"
            +
            "# " + node.getName() + "\n"
            +
            "# -----------------------" + "\n"
            +
            "export JONAS_BASE=" + node.getJonasBase() + "\n"
            +
            "export CATALINA_BASE=$JONAS_BASE" + "\n"
            +
            "rm -f $JONAS_BASE/logs/*" + "\n"
            +
            "cp $JONAS_ROOT/examples/sampleCluster2/output/apps/sampleCluster2.ear $JONAS_BASE/apps/autoload" + "\n"
            +
            "jonas start -win -n " + node.getName() + " -Ddomain.name=" + domainName + "\n";
        return nodeDef;
    }

    /**
     * creation of the unix file
     */
    private void flushUnixScriptFile() {
        JEcho echo = new JEcho();
        echo.setDestDir(unixScriptFile);
        String nodesDefs = "";
        for (Iterator it = this.nodesList.iterator(); it.hasNext();) {
            Node node = (Node) it.next();

            nodesDefs = nodesDefs + getLinuxNodeDef(node);
        }

        String contentFile = "\n"
            +
            "# -----------------------" + "\n"
            +
            "# Start the cluster nodes" + "\n"
            +
            "# Note : JONAS_ROOT has to be set" + "\n"
            +
            "# -----------------------" + "\n"
            +
            nodesDefs + "\n";

        echo.setMessage(contentFile);
        echo.setLogInfo(INFO + "Flushing Configuration in '" + unixScriptFile + "'");
        addTask(echo);
    }

    /**
     * get a node launching script
     * @param node node
     * @return definition of the node
     */
    private String getWinNodeDef(Node node) {

        String jb = node.getJonasBase().replace('/', '\\');

        String nodeDef = "\n"
            +
            ":: -----------------------" + "\n"
            +
            ":: " + node.getName() + "\n"
            +
            ":: -----------------------" + "\n"
            +
            "set JONAS_BASE=" + jb + "\n"
            +
            "set CATALINA_BASE=%JONAS_BASE%" + "\n"
            +
            "cd %JONAS_BASE%\\logs" + "\n"
            +
            "del * /Q" + "\n"
            +
            "copy %JONAS_ROOT%\\examples\\sampleCluster2\\output\\apps\\sampleCluster2.ear %JONAS_BASE%\\apps\\autoload" + "\n"
            +
            "call jonas start -win -n " + node.getName() + " -Ddomain.name=" + domainName + "\n";
        return nodeDef;
    }

    /**
     * creation of the windows file
     */
    private void flushWinScriptFile() {
        JEcho echo = new JEcho();
        echo.setDestDir(winScriptFile);
        String nodesDefs = "";
        for (Iterator it = this.nodesList.iterator(); it.hasNext();) {
            Node node = (Node) it.next();

            nodesDefs = nodesDefs + getWinNodeDef(node);
        }

        String contentFile = "\n"
            +
            ":: -----------------------" + "\n"
            +
            ":: Start the cluster nodes" + "\n"
            +
            ":: Note : JONAS_ROOT has to be set" + "\n"
            +
            ":: -----------------------" + "\n"
            +
            nodesDefs + "\n";

        echo.setMessage(contentFile);
        echo.setLogInfo(INFO + "Flushing Configuration in '" + winScriptFile + "'");
        addTask(echo);
    }

    /**
     * Generation of the config files
     */
    public void flushFiles() {
        flushUnixScriptFile();
        flushWinScriptFile();
    }

    /**
     * Define an inner class for nodes
     * @author Benoit Pelletier
     */
    public class Node {

        /**
         * name
         */
        private String name = null;

        /**
         * get name
         * @return name
         */
        public String getName() {
            return name;
        }

        /**
         * set name
         * @param name name to set
         */
        public void setName(String name) {
            this.name = name;
        }

        /**
         * jonasBase dir
         */
        private String jonasBase = null;

        /**
         * @return JONAS_BASE dir
         */
        public String getJonasBase() {
            return jonasBase;
        }

        /**
         * set JONAS_BASE
         * @param jonasBase dir to set
         */
        public void setJonasBase(String jonasBase) {
            this.jonasBase = jonasBase;
        }
    }


    /**
     * Set the domain name
     * @param domainName domain name
     */
    public void setDomainName(String domainName) {
        this.domainName = domainName;
    }
}