/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.discovery;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Vector;

import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.InvalidAttributeValueException;
import javax.management.JMException;
import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.Attribute;
import javax.management.ReflectionException;
import javax.management.remote.JMXServiceURL;
import javax.naming.Context;
import javax.naming.NamingException;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.J2eeObjectName;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas_domain.api.DomainMap;
import org.objectweb.jonas_domain.api.DomainMapException;
import org.objectweb.jonas_domain.lib.wrapper.DomainManagerWrapper;
import org.objectweb.jonas_domain.xml.Cluster;
import org.objectweb.jonas_domain.xml.Location;
import org.objectweb.jonas_domain.xml.Server;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 *
 * @author Adriana Danes
 *
 * The discovery service creates and starts at least a DiscoveryManager which is a MBean
 * that multicasts <code>discovery information</code> to all the servers who joined the
 * <code>discovery multicast group</code>.
 * <p>
 * The discovery service may also create a Enroller and a DiscoveryClient, and in this case
 * the current server becomes a <code>discovery server</code>.
 * <p>
 * <code>Discovery information</code> contains information allowing to remotely manage a server in the group.
 * <p>
 * <code>Discovery multicast group</code> is a group of servers which can be managed remotely by a discovery server.
 * <p>
 * <code>Discovery server</code> is the server in the group who detains the discovery information concerning all the
 * servers in the group.
 */
public class DiscoveryServiceImpl extends AbsServiceImpl implements DiscoveryService , DiscoveryServiceImplMBean {

    /**
     * The default port where responses to discovery events should be sent.
     */
    private static final String DISCOVERY_SOURCE_PORT_DEFAULT = "9888";
    /**
     * The default port where responses to greeting messages should be sent.
     */
    private static final String DISCOVERY_GREETING_PORT_DEFAULT = "9899";
    /**
     * The default timeout duration that server should wait to receive
     * responses to it's greeting message before continuing.
     */
    private static final String DISCOVERY_GREETING_TIMEOUT_DEFAULT = "1000";
    /**
     * String form of the IP Address where multicast messages are sent.
     */
    private String listeningIp = null;
    /**
     * The port to listen to for multicast messages.
     */
    private int listeningPort;
    /**
     * The port to listen to for responses to greeting message.
     */
    private int greetingListeningPort;
    /**
     * The time period in miliseconds to listen for greeting responses.
     */
    private int greetingAckTimeOut;
    /**
     * The port where discovery events are sent.
     */
    private int sourcePort;
    /**
     * set to true if enroller and discoveryClient started
     */
    private boolean isDiscoveryMaster = false;
    /**
     * Reference to a MBean server.
     */
    private MBeanServer mbeanServer = null;

    /**
     * jmx service reference
     */
    private JmxService jmxService = null;
    /**
     * Logger for this service.
     */
    private static Logger logger = null;
    /**
     * Reference to the Enroller MBean instance
     * The Enrooler can be used by the J2EEDomain MBean to sent a discovery notification to the JMX service
     * in order to add a server to the domain.
     */
    private Enroller enroller = null;

    private ObjectName domainOn = null;

    /**
     * @return the multicast group IP address used by the discovery service
     */
    public String getMulticastAddress() {
        return listeningIp;
    }

    /**
     * @return the multicast group port number used by the discovery service
     */
    public String getMulticastPort() {
        return String.valueOf(listeningPort);
    }

    /**
     * @return true if the current server is a discovery server
     */
    public Boolean getIsDiscoveryMaster() {
        return new Boolean(isDiscoveryMaster);
    }
    /**
     * Management operation allowing to make the current server become a
     * master if its not already.
     * @throws JMException a JMX exception occured when trying to make current server a discovery master
     */
    public void startDiscoveryMaster() throws JMException {
        if (!isDiscoveryMaster) {
            createEnroller();
            createDiscClient();
            isDiscoveryMaster = true;
            setToMaster();
        }
    }

    /**
     * Initialize the discovery service
     * @param ctx context containing initialization parameters
     * @exception discovery service could not be correctly initialized
     */
    protected void doInit(Context ctx) throws ServiceException {
        //Init the logger
        logger = Log.getLogger(Log.JONAS_DISCOVERY_PREFIX);

        // Get service configuration
        try {
            listeningIp = (String) ctx.lookup("jonas.service.discovery.multicast.address");
            String sListeningPort = (String) ctx.lookup("jonas.service.discovery.multicast.port");
            listeningPort = (Integer.valueOf(sListeningPort)).intValue();

        } catch (NamingException ne) {
            String err = "Cannot read initializations arguments in service context";
            logger.log(BasicLevel.ERROR, err);
            throw new ServiceException(err, ne);
        }
        // Get info allowing to see if this is a discovery master
        try {
            String sMaster = (String) ctx.lookup("jonas.service.discovery.master");
            if (sMaster != null && sMaster.equals("true")) {
                isDiscoveryMaster = true;
            }
        } catch (NamingException ne) {
            // isDiscoveryMaster rests false
        }
        // If configuration does not says this ia a master, allow
        // however to start it as a master if the server name is
        // identical to the domain name (in the default case this
        // is "jonas", "jonas")
        if (!isDiscoveryMaster) {
            if (getDomainName().equals(getJonasServerName())) {
                isDiscoveryMaster = true;
            }
        }
        // Temporary strings to store string forms of ports and timeout.
        String sSourcePort = null;
        String greetListeningPort = null;
        String greetAckTimeOut = null;

        // Check if the source port number has been defined else use default
        try {
            sSourcePort = (String) ctx.lookup("jonas.service.discovery.source.port");
        } catch (NamingException ne) {
            sSourcePort = DISCOVERY_SOURCE_PORT_DEFAULT;
        }
        sourcePort = (Integer.valueOf(sSourcePort)).intValue();

        // Check if the greeting port has been defined else use default
        try {
            greetListeningPort = (String) ctx.lookup("jonas.service.discovery.greeting.port");
        } catch (NamingException e1) {
            greetListeningPort = DISCOVERY_GREETING_PORT_DEFAULT;
        }
        greetingListeningPort = (Integer.valueOf(greetListeningPort)).intValue();

        // Check if the greeting timeout has been defined else use default
        try {
            greetAckTimeOut = (String) ctx.lookup("jonas.service.discovery.greeting.timeout");
        } catch (NamingException e1) {
            greetAckTimeOut = DISCOVERY_GREETING_TIMEOUT_DEFAULT;
        }
        greetingAckTimeOut = (Integer.valueOf(greetAckTimeOut)).intValue();

        // Get JOnAS references allowing to start execution
        ServiceManager sm  = null;
        try {
            sm = ServiceManager.getInstance();
        } catch (Exception e) {
            String err = "Cannot get ServiceManager instance";
            logger.log(BasicLevel.ERROR, err);
            throw new ServiceException(err, e);
        }
        jmxService = ((JmxService) sm.getJmxService());
        try {
            mbeanServer = jmxService.getJmxServer();
        } catch (ServiceException e) {
            // the JMX service may not be started
            mbeanServer = null;
            String err = "Cannot get MBeanServer reference";
            logger.log(BasicLevel.ERROR, err);
            throw new ServiceException(err, e);
        }
        domainOn = J2eeObjectName.J2EEDomain(getDomainName());
    }

    /**
     * Start the discovery service
     *
     * @throws ServiceException
     *             An error occured when starting the service
     */
    protected void doStart() throws ServiceException {
        // Create discovery manager
        DiscoveryManager dm = new DiscoveryManager(this.getJonasServerName(), listeningPort,
                listeningIp, greetingListeningPort, greetingAckTimeOut);
        dm.setDomainName(jmxService.getDomainName());
        dm.setJonasName(jmxService.getJonasServerName());
        JMXServiceURL[] connectorServerURLs = jmxService
        .getConnectorServerURLs();
        ArrayList urlsList = new ArrayList();
        for (int i = 0; i < connectorServerURLs.length; i++) {
            // The connectorServerURLs may contain null
            // if the list of protocols in Carol contain
            // other protocols than the standard ones (JRMP, JEREMIE, IIOP, CMI)
            if (connectorServerURLs[i] != null) {
                urlsList.add(connectorServerURLs[i].toString());
            }
        }
        String[] urls = new String[urlsList.size()];
        for (int i = 0; i < urls.length; i++) {
            urls[i] = (String) urlsList.get(i);
        }
        dm.setUrls(urls);
        try {
            // Register DiscoveryManager MBean
            if (mbeanServer != null) {
                mbeanServer.registerMBean(dm, JonasObjectName
                        .discoveryManager());
            }
        } catch (JMException e) {
            throw new ServiceException(
                    "Problem when starting the Discovery Service: ", e);
        }
        // Start discovery manager
        try {
            dm.start();
        } catch (DuplicateServerNameException e) {
            // Server with the same name already exists in domain.
            logger.log(BasicLevel.ERROR,
                    "Discovery manager failed to start due to a pre-existing"
                    + " server in the domain with the same name.", e);
            try {
                // Discovery manager failed to start so remove the mbean.
                mbeanServer.unregisterMBean(JonasObjectName.discoveryManager());
            } catch (JMException e1) {
                logger.log(BasicLevel.ERROR, "Problem trying to unregister "
                        + "DiscoveryManager: ", e);
            }
            // Discovery service had a problem.
            throw new ServiceException(
                    "Problem when starting the Discovery Service:", e);
        }

        if (isDiscoveryMaster) {
            // Create enroller
            try {
                createEnroller();
            } catch (JMException e) {
                throw new ServiceException(
                        "Problem when starting the Discovery Service: ", e);
            }

            // Update the J2EEDomain MBean
            try {
                setToMaster();
            } catch (JMException me) {
                logger.log(BasicLevel.INFO, "Can't set J2EEDomain's master attribute: ", me);
            }

            // Create domain map based on the info contained in domain.xml file.
            // This operaration can be done only after creating the enroller
            // which is needed for sending discovery notifications by JmxService
            // and by the J2EEDomain MBeans.
            try {
                createDomainMap();
            } catch (DomainMapException e) {
                // domain.xml not found
                logger.log(BasicLevel.INFO, "No domain.xml found");
            } catch (JMException me) {
                logger.log(BasicLevel.ERROR, "Problem when creating the domain map: ", me);
            }

            // Create the discovery client
            try {
                createDiscClient();
            } catch (JMException e) {
                throw new ServiceException(
                        "Problem when starting the Discovery Service: ", e);
            }
        }

        // Create and register the service MBean
        try {
            if (mbeanServer != null) {
                mbeanServer.registerMBean(this, JonasObjectName
                        .discoveryService());
            }
        } catch (JMException e) {
            throw new ServiceException(
                    "Problem when starting the Discovery Service: ", e);
        }
    }
    /**
     * Create the Enroller MBean
     * @throws JMException
     */
    private void createEnroller() throws JMException {
        enroller = new Enroller(listeningPort, listeningIp);
        if (mbeanServer != null) {
            mbeanServer.registerMBean(enroller, JonasObjectName.discoveryEnroller());
        }
    }
    /**
     * Create the DiscoveryClient MBean
     * @throws JMException
     */
    private void createDiscClient() throws JMException {
        DiscoveryClient dc = new DiscoveryClient(listeningPort, listeningIp, sourcePort);
        // Register MBean
        if (mbeanServer != null) {
            mbeanServer.registerMBean(dc, JonasObjectName.discoveryClient());
        }
    }
    /**
     * The the J2EEDomain MBeans's master attribute to true
     * @throws JMException
     *
     */
    private void setToMaster() throws JMException {
        Attribute masterAtt = new Attribute("master", new Boolean(true));
        mbeanServer.setAttribute(domainOn, masterAtt);
    }
    /**
     * Create the domain representation by adding server (J2EEServer) and cluster (J2EEDomain) MBeans
     * corresponding to information read in the domain.xml file.
     * @throws DomainMapException
     * @throws JMException
     */
    private void createDomainMap() throws DomainMapException, JMException {
        String domainName = getDomainName();
        ClassLoader currentLoader = Thread.currentThread().getContextClassLoader();
        DomainMap domainMap = DomainManagerWrapper.getDomainMap(null, currentLoader);
        String definedName =  domainMap.getName();
        if (domainName.equals(definedName)) {
            Attribute descriptionAtt = new Attribute("description", domainMap.getDescription());
            mbeanServer.setAttribute(domainOn, descriptionAtt);
            Vector servers = domainMap.getServers();
            if (servers != null) {
                for (int i = 0; i < servers.size(); i++) {
                    Server serverElement = (Server) servers.elementAt(i);
                    addServer(domainOn, domainName, serverElement);
                }
            }
            Vector clusters = domainMap.getClusters();
            if (clusters != null) {
                for (int i = 0; i < clusters.size(); i++) {
                    Cluster clusterElement = (Cluster) clusters.elementAt(i);
                    addCluster(domainOn, domainName, clusterElement);
                }
            }
        } else {
            logger.log(BasicLevel.INFO, "Domain name in domain.xml is not the actual domain name: " + domainName);
        }
    }
    /**
     * Create a J2EEServer MBean and add it to the target's servers list
     * @param target a domain or cluster MBean
     * @param domainName the name of the domain
     * @param serverElement containes infos abour the server to add
     * @throws JMException when the J2EEServer MBean could not be added to the target
     */
    private void addServer(ObjectName target, String domainName, Server serverElement) throws JMException {
        String targetName = target.getKeyProperty("name");
        String clusterName = target.getKeyProperty("clusterName");
        if (clusterName != null) {
            targetName = clusterName;
        }
        String serverName = serverElement.getName();
        Location serverLocation = (Location) serverElement.getLocation();
        LinkedList urlList = serverLocation.getUrlList();
        String[] urls = new String[urlList.size()];
        for (int j = 0; j < urlList.size(); j++) {
            String url = (String) urlList.get(j);
            if (url != null) {
                urls[j] = url;
            }
        }
        String[] signature = new String[2];
        Object[] params = new Object[2];
        signature[0] = "java.lang.String";
        signature[1] = "[Ljava.lang.String;";
        ObjectName serverOn = J2eeObjectName.J2EEServer(domainName, serverName);
        params[0] = serverOn.toString();
        params[1] = urls;
        mbeanServer.invoke(target, "addServer", params, signature);
    }
    /**
     * Create a J2EEDomain MBean and add it to the target's clusters list
     * @param target a domain or cluster MBean
     * @param domainName the name of the domain
     * @param clusterElement containes infos abour the cluster to add
     * @throws JMException when the J2EECluster MBean could not be added to the target
     */
    private void addCluster(ObjectName target, String domainName, Cluster clusterElement) throws JMException {
        String targetName = target.getKeyProperty("name");
        String targetClusterName = target.getKeyProperty("clusterName");
        if (targetClusterName != null) {
            targetName = targetClusterName;
        }
        String clusterName = clusterElement.getName();
        String[] signature = new String[1];
        signature[0] = "java.lang.String";
        Object[] params = new Object[1];
        params[0] = clusterName;
        String clusterON = (String) mbeanServer.invoke(target, "createCluster", params, signature);
        ObjectName clusterOn = ObjectName.getInstance(clusterON);
        LinkedList serversList = clusterElement.getServerList();
        for (int j = 0; j < serversList.size(); j++) {
            Server serverElement = (Server) serversList.get(j);
            addServer(clusterOn, domainName, serverElement);
        }
        LinkedList clustersList = clusterElement.getClusterList();
        for (int j = 0; j < clustersList.size(); j++) {
            Cluster innerClusterElement = (Cluster) clustersList.get(j);
            addCluster(clusterOn, domainName, innerClusterElement);
        }
    }

    protected void doStop() throws ServiceException {
        // TO DO
        // ....

        // Unegister the service MBean
        try {
            if (mbeanServer != null) {
                mbeanServer.unregisterMBean(JonasObjectName.discoveryService());
            }
        } catch (JMException e) {
            throw new ServiceException("Problem when stoping the Discovery Service: ", e);
        }
    }

    /**
     * @return Returns the enroller.
     */
    public Enroller getEnroller() {
        return enroller;
    }

    /**
     * Create a 'fake' DiscEvent object containing info to establish a JMX connection with a new server
     * in the domain
     * @param serverName the OBJECT_NAME of the server MBean corresponding to the server to connect
     * @param domainName the JOnAS server's domain name
     * @param connectorURLs the urls of the connector server of the server to connect
     * @param state the state of the server (RUNNING in case the DiscEvent is used to add a server; could be STOPPING
     * if the DiscEvent is used to remove a server)
     * @return a new DiscEvent object
     */
    public DiscEvent getDiscEvent(String serverName, String domainName, String[] connectorURLs, String state) {
        String sourceAddress = null;
        int sourcePort = 0;
        String serverId = null;
        DiscEvent fakeMessage = new DiscEvent(sourceAddress, sourcePort, serverName, domainName, serverId, connectorURLs);
        fakeMessage.setState(state);
        return fakeMessage;
    }
    /**
     * @return the discovery protocol version number
     */
     public String getDiscoveryProtocolVersion() {
         return DiscMessage.DISCOVERY_PROTOCOL_VERSION;
     }
}