/**
 * High Availability Service (HA) for JOnAS
 *
 * Copyright (C) 2006 Distributed Systems Lab.
 * Universidad Politécnica de Madrid (Spain)
 * Contact: http://lsd.ls.fi.upm.es/lsd
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.ha.interceptor;

import org.objectweb.carol.cmi.ha.RequestId;
import org.objectweb.carol.cmi.ha.interceptor.HACurrentDelegate;

/**
 * This class handles the HAContext/Thread association
 * @author Francisco Perez-Sorrosal (fpsorrosal@no-spam@fi.upm.es)
 * @author Alberto Paz-Jimenez (apaz@no-spam@fi.upm.es)
 */
public class HACurrentDelegateImpl implements HACurrentDelegate {

    /**
     * Local thread
     */
    private static InheritableThreadLocal threadCtx;

    /**
     * Init the thread
     */
    static {
        threadCtx = new InheritableThreadLocal();
    }

    /**
     * Unique instance
     */
    private static HACurrentDelegateImpl current = new HACurrentDelegateImpl();

    /**
     * Returns the unique instance
     * @return the current instance
     */
    public static HACurrentDelegateImpl getCurrent() {
        return current;
    }

    /**
     * Associates a HA context with the current thread
     * @param ctx HAcontext to associate to the current thread
     */
    public void setHAContext(HAContext ctx) {
        threadCtx.set(ctx);
    }

    /**
     * Returns the HA context associated with the current thread
     * @return the HA context associated to the current thread
     */
    public HAContext getHAContext() {
        return (HAContext) threadCtx.get();
    }

    /**
     * Set the current request in the current context if exists
     * @param currentReq the current request
     */
    public void setCurrentReq(RequestId currentReq) {
        HACurrentContext current = (HACurrentContext) threadCtx.get();
        if (current == null) {
            current = new HACurrentContext();
        }
        current.setCurrentReq(currentReq);
        threadCtx.set(current);
    }

    /**
     * Get the current request in the current context if exists
     * @return the current request
     */
    public RequestId getCurrentReq() {
        HACurrentContext current = (HACurrentContext) threadCtx.get();
        if (current == null) {
            current = new HACurrentContext();
        }
        return current.getCurrentReq();
    }

    /**
     * Set the next request in the current context if exists
     * @param nextReq the next request
     */
    public void setNextReq(RequestId nextReq) {
        HACurrentContext current = (HACurrentContext) threadCtx.get();
        if (current == null) {
            current = new HACurrentContext();
        }
        current.setNextReq(nextReq);
        threadCtx.set(current);
    }

    /**
     * Get the next request in the current context if exists
     * @return the next request
     */
    public RequestId getNextReq() {
        HACurrentContext current = (HACurrentContext) threadCtx.get();
        if (current == null) {
            current = new HACurrentContext();
        }
        return current.getNextReq();
    }

    /**
     * Set the root request in the current context if exists
     * @param nextReq the next request
     */
    public void setRootReq(RequestId nextReq) {
        HACurrentContext current = (HACurrentContext) threadCtx.get();
        if (current == null) {
            current = new HACurrentContext();
        }
        current.setRootReq(nextReq);
        threadCtx.set(current);
    }

    /**
     * Get the root request in the current context if exists
     * @return the root request
     */
    public RequestId getRootReq() {
        HACurrentContext current = (HACurrentContext) threadCtx.get();
        if (current == null) {
            current = new HACurrentContext();
        }
        return current.getRootReq();
    }

    /**
     * Set the if it is necessary to replicate in the current context if exists
     * @param replicate true if it is necessary the replication procedure
     */
    public void setReplicate(boolean replicate) {
        HACurrentContext current = (HACurrentContext) threadCtx.get();
        if (current == null) {
            current = new HACurrentContext();
        }
        current.setReplicate(replicate);
        threadCtx.set(current);
    }

    /**
     * Get the if it is necessary to replicate in the current context if exists
     * @return true if it is necessary the replication procedure
     */
    public boolean replicate() {
        HACurrentContext current = (HACurrentContext) threadCtx.get();
        if (current == null) {
            current = new HACurrentContext();
        }
        return current.isReplicate();
    }

}