/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.management.j2eemanagement;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.management.InstanceNotFoundException;
import javax.management.JMException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.MBeanServerConnection;
import javax.management.MBeanServerNotification;
import javax.management.MalformedObjectNameException;
import javax.management.Notification;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.modelmbean.InvalidTargetObjectTypeException;
import javax.management.modelmbean.ModelMBean;
import javax.management.remote.JMXConnector;

import org.apache.commons.modeler.ManagedBean;
import org.apache.commons.modeler.Registry;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.discovery.DiscEvent;
import org.objectweb.jonas.discovery.DiscoveryListener;
import org.objectweb.jonas.discovery.DiscoveryServiceImpl;
import org.objectweb.jonas.discovery.DiscoveryState;
import org.objectweb.jonas.discovery.Enroller;
import org.objectweb.jonas.jmx.J2eeObjectName;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.management.JonasMBeanTools;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Implements a managed object which represents a management domain.
 * Domain management information is present in the folowwing data structures:
 * <ul>
 * <li><code>myServers</code> containes the OBJECT_NAMEs of the J2EEServers in the domain</li>
 * <li><code>managedServersToUrls</code> containes the JMX Connector URLs
 * each server provides for management</li>
 * <li><code>myCluster</code>containes the OBJECT_NAMEs of the J2EEDomains this domain contains</li>
 * </ul>
 * They are updated by addServer(), removeServer(), createCluster() methods.
 * @author Adriana Danes
 * @author Vivek Lakshmanan
 */
public class J2EEDomain extends J2EEManagedObject implements MBeanRegistration, NotificationListener {

    private static final int BUFFER_SIZE = 1024;

    public static final String STARTED = "started";

    public static final String STOPPED = "stopped";
    /**
     * Management domain name
     */
    private String name = null;
    /**
     * Domain description
     */
    private String description = null;
    /**
     * List of OBJECT_NAMEs corresponding to the servers belonging to this domain (or cluster)
     */
    private List myServers = null;
    /**
     * The 'created' servers are those that are added to the domain by an explicite management operation
     * or at the discovery service start-up whe reading the 'domain.xml' file
     */
    private List myCreatedServers = null;

    /**
     * List of OBJECT_NAMEs corresponding to the clusters defined in this domain (or cluster)
     */
    private List myClusters = null;

    /**
     * List of objects corresponding to the clusters in this domain (or cluster)
     */
    private List myClusterObjects = null;

    /**
     * J2EE Management logger
     */
    private static Logger logger = null;

    /**
     * Current MBeanServer (in which the current J2EEDomain MBean is registered)
     */
    private MBeanServer mbeanServer = null;

    /**
     * JmxService reference allows for getting a connection for a server in the domain
     */
    private JmxService jmxService = null;

    /**
     * Mapping serverName -> JMX Service URLs published by the server named serverName
     */
    private Map managedServersToUrls = null;

    /**
     * Mapping serverName -> serverState
     */
    private Map managedServersToState = null;

    /**
     * Used to construct JMX notifications
     */
    private long sequenceNumber      = 0;

    /**
     * True if this is a cluster, false if this is a domain (the root all the
     * clusters)
     */
    private boolean isCluster = false;
    /**
     * true if the current server is a master
     */
    private boolean master = false;

    private ArrayList registeredOnsForListner = null;

    /**
     * MBean constructor
     * @param objectName object name of the managed object
     * @param stateManageable if true, this managed object implements J2EE State Management Model
     * @param statisticsProvider if true, this managed object implements the J2EE StatisticProvide Model
     * @param eventProvider  if true, this managed object implements the J2EE EventProvider Model
     * @throws JMException ..
     */
    public J2EEDomain(String objectName, boolean stateManageable, boolean statisticsProvider, boolean eventProvider) throws JMException {
        super(objectName, stateManageable, statisticsProvider, eventProvider);
        ObjectName myObjectName = ObjectName.getInstance(objectName);
        name = myObjectName.getKeyProperty("name");
        if (myObjectName.getKeyProperty("clusterName") != null) {
            isCluster = true;
        }
        myServers = Collections.synchronizedList(new ArrayList());
        myCreatedServers = Collections.synchronizedList(new ArrayList());
        myClusters = Collections.synchronizedList(new ArrayList());
        myClusterObjects = Collections.synchronizedList(new ArrayList());
        managedServersToUrls = new HashMap();
        managedServersToState = new HashMap();
        logger = Log.getLogger("org.objectweb.jonas.management.j2eemanagement");
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "J2EEDomain managed object created. OBJECT_NAME = " + objectName);
        }
        registeredOnsForListner = new ArrayList(2);
    }

    /**
     * Return the MBeans OBJECT_NAMEs of the servers belonging to this domain (or cluster)
     * Convert myServers ArrayList of OBJECT_NAMEs to an array of Strings of OBJECT_NAMEs
     * @return The list of object names corresponding to the servers associated with this domain
     */
    public String[] getServers() {
        int nbSubServers = 0;
        int nbClusters = myClusterObjects.size();
        Object[] mySubServers = new Object[nbClusters];
        for (int i = 0; i < nbClusters; i++) {
            J2EEDomain j2eeCluster = (J2EEDomain) myClusterObjects.get(i);
            String[] servers = j2eeCluster.getServers();
            mySubServers[i] = servers;
            nbSubServers = nbSubServers + servers.length;
        }
        // Array for all the servers
        String[] sb = new String[myServers.size() + nbSubServers];
        // Add servers belonging to this domain (cluster)
        int k = 0;
        for (Iterator it = myServers.iterator(); it.hasNext();) {
            sb[k++] = (String) it.next();
            logger.log(BasicLevel.DEBUG, "Got server " + sb[k-1]);
        }
        // Add servers belonging to the sub clusters
        for (int i = 0; i < nbClusters; i++) {
            String[] subServers = (String[]) mySubServers[i];
            for (int j = 0; j < subServers.length; j++) {
                sb[k++] = subServers[j];
                logger.log(BasicLevel.DEBUG, "Got server " + sb[k-1]);
            }
        }
        return sb;
    }

    /**
     * Return the MBeans OBJECT_NAMEs of the clusters in this domain (or cluster).
     * Convert myClusters ArrayList of OBJECT_NAMEs to an array of Strings of OBJECT_NAMEs
     * @return The list of object names corresponding to the clusters in this domain
     */
    public String[] getClusters() {
        //return (String[]) myClusters.toArray();
        String[] sb = new String[myClusters.size()];
        int i = 0;
        for (Iterator it = myClusters.iterator(); it.hasNext();) {
            sb[i++] = (String) it.next();
        }
        return sb;
    }

    /**
     * Returns the names of the servers currently managed.
     *
     * @return Names of servers currently managed.
     */
    public String[] getServerNames() {
        return (String[]) managedServersToUrls.keySet().toArray(new String[managedServersToUrls.keySet().size()]);
    }

    /**
     * Return the connector server urls for a given server.
     * @param serverName the server we are inquiring for
     * @return The list of connector server urls corresponding to the given server
     */
    public String[] getConnectorServerURLs(String serverName) {
        return (String[]) managedServersToUrls.get(serverName);
    }

    /**
     * Return the state of a server belonging to this domain (or cluster)
     * @param serverName the server we are inquiring for
     * @return the state of the server
     */
    public String getServerState(String serverName) {
        return (String) managedServersToState.get(serverName);
    }
    /**
     * Return the names of the STARTED servers
     * @return Array of STARTED server names
     */
    public String[] getStartedServers() {
        ArrayList al = new ArrayList();
        String[] servers = getServerNames();
        for (int i = 0; i < servers.length; i++) {
            if (STARTED.equals((String) managedServersToState.get(servers[i]))) {
                al.add(servers[i]);
            }
        }
        String[] result = new String[al.size()];
        for (int i = 0; i < al.size(); i++) {
            result[i] = (String) al.get(i);
        }
        return result;
    }
    /**
     * Create a J2EEServer MBean corresponding and add its OBJECT_NAME to the <code>servers</code> list.
     * If this method is not called by a handleNotification, use the Enroller MBean to send a discovery notification
     * towards the JMX service in order to create a connection.
     * @param serverOn ObjectName (String form) of a J2EEServer to be add in this domain (or cluster)
     * @param connectorServerURLs connector server urls for this server
     * @param handleNotification true if the method is called by a handleNotification method, false otherwise
     */
    public void addServer(String serverOn, String[] connectorServerURLs, String state, boolean handleNotification) throws MalformedObjectNameException, Exception {

        ObjectName on = ObjectName.getInstance(serverOn);
        String serverName = on.getKeyProperty("name");
        if (isServerInServersList(serverName, getServerNames())) {
            // known server
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "The object name: " + serverOn + " is already in the 'serverNames' list of domain " + name);
            }
            // test if state changed from 'stopped' to 'started'
            String knownState = (String) managedServersToState.get(serverName);
            if (!state.equals(knownState)) {
                // State change
                managedServersToState.put(serverName, state);
                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "Server " + serverName + " state changed from " + knownState + " to " + state);
                }
            }
        } else {
            // unknown server (new server)
            boolean update = false;
            if (state.equals(STOPPED)) {
                update = true; // update if explicite creation
            } else if (!isCluster) {
                // update the root domain as servers started in the domain will by default go
                // into the domain and not in sub-domains (clusters)
                if (!isServerInServersList(serverOn, getServers())) {
                    // but only if this server
                    // is not already in the domain's clusters
                    update = true;
                }
            }
            if (update) {
                myServers.add(serverOn);
                managedServersToUrls.put(serverName, connectorServerURLs);
                managedServersToState.put(serverName, state);
                if (!handleNotification) {
                    // This method was not called as a reaction to a discovery message
                    // get the name of the master server which created this MBean
                    String currentServerName = jmxService.getJonasServerName();
                    if(!currentServerName.equals(serverName)) {
                        // The server to add is not the current server, so this operation
                        // was invoked by an administrator. Send a JMX notification.
                        // We need the discovery service
                        DiscoveryServiceImpl discService = (DiscoveryServiceImpl) ServiceManager.getInstance().getDiscoveryService();
                        // Create a fake discovery event
                        DiscEvent discEvent = discService.getDiscEvent(serverName, name, connectorServerURLs, state);
                        Notification notif = new Notification(DiscoveryListener.DISCOVERY_TYPE, this,
                                sequenceNumber++, System.currentTimeMillis(), state);
                        notif.setUserData(discEvent);
                        // get the Enroller from the discovery service
                        Enroller enroller = discService.getEnroller();
                        enroller.sendNotification(notif);
                    }
                }
            }
        }
    }
    /**
     * Create a J2EEServer MBean and add its OBJECT_NAME to the <code>servers</code> list.
     * This method is exposed as a management operation by the J2EEDomain MBean and can be used
     * to add servers to a domain (or cluster) by a management application or by the entity who reads
     * the discovery.xml file (currently the discovery service).
     * @param serverOn ObjectName (String form) of a J2EEServer to be add in this domain (or cluster)
     * @param connectorServerURLs connector server urls for this server
     * @param handleNotification true if the method is called by a handleNotification method, false otherwise
     */
    public void addServer(String serverOn, String[] connectorServerURLs) throws MalformedObjectNameException, Exception  {
        String state = STOPPED;
        myCreatedServers.add(serverOn);
        this.addServer(serverOn, connectorServerURLs, state, false);
    }
    /**
     * Create a J2EEDomain MBean corresponding to a cluster and add it to the <code>clusters</code> list.
     * @param clusterName the cluster's name
     * @return the OBJECT_NAME corresponding to the created MBean
     */
    public String createCluster(String clusterName) {
        String result = null;
        // Determine the parent of the cluster
        String parentClusterName = null;
        try {
            // This MBean's ObjectName
            ObjectName myOn = ObjectName.getInstance(getObjectName());
            parentClusterName = myOn.getKeyProperty("clusterName");
            if (parentClusterName == null) {
                // this is a root cluster (its parent is the domain)
                parentClusterName = myOn.getDomain();
            }
        } catch (MalformedObjectNameException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        }
        // Create a J2EEDomain ObjectName for the cluster
        ObjectName clusterOn = J2eeObjectName.J2EECluster(name, clusterName, parentClusterName);
        result = clusterOn.toString();
        if (myClusters.contains(clusterOn.toString())) {
            logger.log(BasicLevel.DEBUG, "The cluster " + clusterName + " already exists in " + parentClusterName);
        } else {
            try {
                // Get the Registry MBean and load mbeans-descriptors
                Registry oRegistry = JonasMBeanTools.getRegistry();
                // Create the J2EEDomain MBean instance which is the MBean associated to the cluster
                J2EEDomain j2eeCluster = new J2EEDomain(clusterOn.toString(), false, false, false);
                // Set master attribute to true for JonasAdmin in order to allow cluster management
                j2eeCluster.setMaster(true);
                // update registered MBeans for notification listner
                j2eeCluster.setRegisteredOnsForListner(getRegisteredOnsForListner());
                // Create managed MBean for the cluster MBean and register it
                ManagedBean oManaged = oRegistry.findManagedBean("J2EEDomain");
                ModelMBean oMBean = oManaged.createMBean(j2eeCluster);
                mbeanServer.registerMBean(oMBean, clusterOn);
                // Update j2eeCluster
                j2eeCluster.setJmxService(jmxService);
                // Put it into myClusterObjects list
                myClusterObjects.add(j2eeCluster);
            } catch (JMException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            } catch (InvalidTargetObjectTypeException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
            // update the clusters list
            myClusters.add(clusterOn.toString());
        }
        return result;
    }

    /**
     * Remove an object name from the <code>servers</code> list.
     *
     * @param serverOn
     *            ObjectName (String form) of a J2EEServer to be removed from
     *            this domain
     * @return Object name of the removed J2EEServer
     */
    public String removeServer(String serverOn) {
        int index = myServers.indexOf(serverOn);
        if (index == -1) {
            // name not found
            return null;
        } else {
            String name = null;
            try {
                name = ObjectName.getInstance(serverOn).getKeyProperty("name");
            } catch (MalformedObjectNameException me) {
                // No mapping for this server name, this shouldnt happen but if
                // it does, return null and don't remove anything.
                return null;
            }
            boolean doRemove = false;
            if (myCreatedServers.contains(serverOn)) {
                // This server was add to the domain/cluster by a direct call to the addServer
                // method (it could be defined in domain.xml file).
                String knownState = (String) managedServersToState.get(name);
                if (STARTED.equals(knownState)) {
                    managedServersToState.put(name, STOPPED);
                } else if (STOPPED.equals(knownState)){
                    doRemove = true;
                }
            } else {
                // This server was add to the domain by the discovery services
                doRemove = true;
            }
            if (doRemove) {
                managedServersToUrls.remove(name);
                managedServersToState.remove(name);
                return (String) myServers.remove(index);
            } else {
                return null;
            }
        }
    }

    /**
     * @param server The MBean server in which the MBean will be registered.
     * @param name The object name of the MBean
     * @return The object name of the MBean
     */
    public ObjectName preRegister(MBeanServer server, ObjectName name) throws Exception {
        if (name == null) {
            return J2eeObjectName.J2EEDomain(this.name);
        }
        String myName = getObjectName().toString();
        String argName = name.toString();
        if (myName.equals(argName)) {
            // OK
            mbeanServer = server;
            return name;
        } else {
            if (logger.isLoggable(BasicLevel.WARN)) {
                logger.log(BasicLevel.WARN, "Trying to register J2EEDomain MBean with the following name: "
                        + argName + " when the expected name is: " + myName);
            }
            return ObjectName.getInstance(this.name);
        }
    }

    /**
     * Add J2EEDomain MBean (myself) as listener to registration/unregistration notifications of
     * JOnAS management MBeans.
     * @param registrationDone Indicates whether or not the MBean has been successfully registered
     */
    public void postRegister(Boolean registrationDone) {
        if (registrationDone.booleanValue()) {
            if (isCluster()) {
                // Register as listenet to notification sent by MBeans in registeredOnsForListner
                // (Enroller or DiscoveryClient MBeans)
                for (int i = 0; i < registeredOnsForListner.size(); i++) {
                    try {
                        mbeanServer.addNotificationListener((ObjectName) registeredOnsForListner.get(i), this, null, null);
                        if (logger.isLoggable(BasicLevel.DEBUG)) {
                            logger.log(BasicLevel.DEBUG, "J2EEDomain (this) registered as listener to notifs emitted by " + (ObjectName) registeredOnsForListner.get(i));
                        }
                    } catch (InstanceNotFoundException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }
            }
            try {
                // MBeanServerDelegate is the MBean which sends registration/unregistration notifications
                ObjectName delegate = ObjectName.getInstance("JMImplementation:type=MBeanServerDelegate");
                mbeanServer.addNotificationListener(delegate, this, null, null);
            } catch (JMException me) {
                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "J2EEDomain MBean could not be added as notification listener for MBeanServerNotifications "
                            + "related to the REGISTRATION or UNREGISTRETION of JOnAS management MBeans");
                }
            }
        }
    }

    /**
     * @see javax.management.MBeanRegistration#preDeregister()
     */
    public void preDeregister() throws Exception {
        // TODO Auto-generated method stub

    }

    /**
     * @see javax.management.MBeanRegistration#postDeregister()
     */
    public void postDeregister() {
        // TODO Auto-generated method stub

    }

    /**
     * @param notification received notification
     * @param handback received handback
     */
    public void handleNotification(Notification notification, Object handback) {
        if (notification instanceof MBeanServerNotification) {
            handleMBeanServerNotification((MBeanServerNotification) notification);
        } else {
            // Handle notification emmited by the Enroller or the DiscoveryClient MBean
            handleDiscoveryNotification(notification);
        }
    }

    /**
     * Treat REGISTRATION/UNREGISTRATION MBeanServerNotification generated for "management" MBeans
     * @param notification MBeanServerNotification to treat
     */
    private void handleMBeanServerNotification(MBeanServerNotification notification) {
        String type = notification.getType();
        // ObjectName of the MBean that caused the notification
        ObjectName registeredOn = ((MBeanServerNotification) notification).getMBeanName();
        String name = registeredOn.getKeyProperty("name");
        // The names below are defined in JonasObjectName class
        if ("discoveryEnroller".equals(name) || "discoveryClient".equals(name)) {
            if (type.equals(MBeanServerNotification.REGISTRATION_NOTIFICATION)) {
                try {
                    // register myself as listener of notifications emitted by the Enroller or DiscoveryClient MBean
                    mbeanServer.addNotificationListener(registeredOn, this, null, null);
                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG, "J2EEDomain (this) registered as listener to notifs emitted by " + registeredOn);
                    }
                    registeredOnsForListner.add(registeredOn);
                } catch (InstanceNotFoundException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
            if (type.equals(MBeanServerNotification.UNREGISTRATION_NOTIFICATION)) {
                try {
                    // unregister myself as listener of notifications emitted by the Enroller or DiscoveryClient MBean
                    mbeanServer.removeNotificationListener(registeredOn, this);
                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG, "J2EEDomain (this) removed listener for notifs emitted by " + registeredOn);
                    }
                } catch (InstanceNotFoundException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                } catch (ListenerNotFoundException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
        }
    }

    /**
     * Treat notification if it is a discovery notification
     * @param notification received notification to treat
     */
    private void handleDiscoveryNotification(Notification notification) {
        // Treat discovery notification
        String type = notification.getType();
        if (type.equals(DiscoveryListener.DISCOVERY_TYPE)) {
            DiscEvent userData = (DiscEvent) notification.getUserData();
            String domainName = userData.getDomainName();
            if (!domainName.equals(name)) {
            	return;
            }
            String serverName = userData.getServerName();
            ObjectName on = J2eeObjectName.J2EEServer(domainName, serverName);
            String state = userData.getState();
            if (state.equals(DiscoveryState.RUNNING)) {
                try {
                    addServer(on.toString(), userData.getConnectorURL(), STARTED, true);
                } catch (MalformedObjectNameException moe) {
                    // Normally this is not possible, on was created below
                    moe.printStackTrace();
                } catch (Exception e) {
                    // if could not access to the JOnAS ServiceManager to get ref on the discovery service
                    e.printStackTrace();
                }
            } else if (state.equals(DiscoveryState.STOPPING)) {
                removeServer(on.toString());
            }
        }
    }

    // -------------------------------------
    /**
     * Support of management operations at domain level.
     * --> J2EE Application and modules deployment
     * -------------------------------------------------
     * In this implementation only one deployment operation can be launched
     * at a time. During the duration of this operation, the value of the
     * <code>progress</code> object is not null. A new operation request
     * will be rejected if the the value of the <code>progress</code> object is
     * RUNNING.
     */
    static final private String RUNNING = "running";
    static final private String COMPLETED = "completed";
    static final private String FAILED = "failed";
    static final private String DEPLOYED = "deployed";
    static final private String UNDEPLOYED = "undeployed";

    static final private String DEPLOY = "deploy";
    static final private String UNDEPLOY = "undeploy";
    static final private String UPLOAD_DEPLOY = "uploadDeploy";
    static final private String UPLOAD = "upload";

    static final private boolean SUCCESS = true;
    static final private boolean FAILURE = false;

    private String progress = null;
    private HashMap progressReport = null;
    private String globalErrorReport = null;

    /**
     * @return Returns the progress value.
     */
    public String getProgress() {
        return progress;
    }

    /**
     * @param progress The progress to set.
     */
    private void setProgress(String progress) {
        this.progress = progress;
    }

    /**
     * Set progress to null
     */
    public void resetProgress() {
        progress = null;
    }

    /**
     * Deploy a JAR module on a target if no other deployment operation is executing.
     *
     * @param target The OBJECT_NAMES of the J2EEServers composing the target
     * @param fileName file containing the module (installed on each target)
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean deployJar(String[] target, String fileName) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "deployJar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, DEPLOY, false);
    }

    /**
     * Un-deploy a JAR module on a target if no other deployment operation is executing.
     *
     * @param target The OBJECT_NAMES of the J2EEServers composing the target
     * @param fileName file containing the module
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean unDeployJar(String[] target, String fileName) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "unDeployJar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, UNDEPLOY, false);
    }

    /**
     * Deploy a WAR module on a target if no other deployment operation is executing.
     *
     * @param target The OBJECT_NAMES of the J2EEServers composing the target
     * @param fileName file containing the module (installed on each target)
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean deployWar(String[] target, String fileName) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "deployWar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, DEPLOY, false);
    }

    /**
     * Un-deploy a WAR module on a target if no other deployment operation is executing.
     *
     * @param target The OBJECT_NAMES of the J2EEServers composing the target
     * @param fileName file containing the module
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean unDeployWar(String[] target, String fileName) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "unDeployWar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, UNDEPLOY, false);
    }

    /**
     * Deploy a RAR module on a target if no other deployment operation is executing.
     *
     * @param target The OBJECT_NAMES of the J2EEServers composing the target
     * @param fileName file containing the module (installed on each target)
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean deployRar(String[] target, String fileName) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "deployRar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, DEPLOY, false);
    }

    /**
     * Un-deploy a RAR module on a target if no other deployment operation is executing.
     *
     * @param target The OBJECT_NAMES of the J2EEServers composing the target
     * @param fileName file containing the module
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean unDeployRar(String[] target, String fileName) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "unDeployRar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, UNDEPLOY, false);
    }

    /**
     * Deploy a EAR module on a target if no other deployment operation is executing.
     *
     * @param target The OBJECT_NAMES of the J2EEServers composing the target
     * @param fileName file containing the module (installed on each target)
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean deployEar(String[] target, String fileName) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "deployEar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, DEPLOY, false);
    }

    /**
     * Un-deploy a EAR module on a target if no other deployment operation is executing.
     *
     * @param target The OBJECT_NAMES of the J2EEServers composing the target
     * @param fileName file containing the module
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean unDeployEar(String[] target, String fileName) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "unDeployEar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, UNDEPLOY, false);
    }

    /**
     * Upload an EAR modlue to all targets and deploy the new module.
     *
     * @param target a list target servers
     * @param fileName file containing the module
     * @param replaceExisting should a file with the same name on the target server be overwritten
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean uploadDeployEar(String[] target, String fileName, boolean replaceExisting) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "reDeployEar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, UPLOAD_DEPLOY, replaceExisting);
    }

    /**
     * Upload an WAR modlue to all targets and deploy the new module.
     *
     * @param target a list target servers
     * @param fileName file containing the module
     * @param replaceExisting should a file with the same name on the target server be overwritten
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean uploadDeployWar(String[] target, String fileName, boolean replaceExisting) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "reDeployWar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, UPLOAD_DEPLOY, replaceExisting);
    }

    /**
     * Upload an JAR modlue to all targets and deploy the new module.
     *
     * @param target a list target servers
     * @param fileName file containing the module
     * @param replaceExisting should a file with the same name on the target server be overwritten
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean uploadDeployJar(String[] target, String fileName, boolean replaceExisting) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "reDeployJar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, UPLOAD_DEPLOY, replaceExisting);
    }

    /**
     * Upload an RAR modlue to all targets and deploy the new module.
     *
     * @param target a list target servers
     * @param fileName file containing the module
     * @param replaceExisting should a file with the same name on the target server be overwritten
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean uploadDeployRar(String[] target, String fileName, boolean replaceExisting) {
        if (isAnotherOperationRunning()) {
            return false;
        }

        String opName = "reDeployRar";
        String[] params = {fileName };
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, UPLOAD_DEPLOY, replaceExisting);
    }

    /**
     * Upload a module to each of the target servers.
     *
     * @param target a list of the target servers
     * @param fileName file containing the module
     * @param replaceExisting should a file with the same name on the target server be overwritten
     * @return true if the operation was executed, false if the operation was rejected
     */
    public boolean uploadFile(String[] target, String fileName, boolean replaceExisting) {
        if (isAnotherOperationRunning()) {
            return false;
        };

        String opName = "uploadFile";
        String[] params = {fileName};
        String[] signature = {"java.lang.String" };
        return doDeploymentOperation(opName, params, signature, target, fileName, UPLOAD, replaceExisting);
    }

    /**
     * A deployment operation can be: deploy/unDeploy/upload/uploadDeploy for
     * jar, war, rar or ear.
     *
     * The progress object reflects the operation evolution.
     *
     * If the operation succeded on every server in the target, the progress passes
     * to COMPLETED. If the op. failed on some servers in the target, the error
     * is reported in the corresponding entry of progressReport HashMap.
     *
     * If the target is not correct (does not correspond to the domain management
     * information), the globalErrorReport contains corresponding error description.
     *
     * @param opName the name of the operation to be invoked
     * @param opParams the parameters to be given to the invoke() method
     * @param opSignature the signature of the method being invoked
     * @param target The OBJECT_NAMES of the J2EEServers composing the target
     * @param fileName file containing the module
     * @param opType one of DEPLOY, UNDEPLOY, UPLOAD, UPLOAD_DEPLOY
     * @param replaceExisting whether an existing module should be overwritten on upload
     *
     * @return true if the operation was executed, false if the operation was
     * rejected
     */
    private boolean doDeploymentOperation(String opName, Object[] opParams, String[] opSignature, String[] target, String fileName, String opType, boolean replaceExisting) {
        progressReport = new HashMap();
        // test target coherence
        if (!isTargetCorrect(target)) {
                setProgress(FAILED);
                return false;
        }
        MBeanServerConnection connection = null;
        JMXConnector connector = null;
        for (int i = 0; i < target.length; i++) {
            ObjectName targetOn = null;
            try {
                targetOn = ObjectName.getInstance(target[i]);
            } catch (MalformedObjectNameException me) {
                // Should not occur here
                setProgress(FAILED);
                continue;
            }
            String serverName = targetOn.getKeyProperty("name");
            connection = jmxService.getServerConnection(serverName);
            try {
                testConnection(connection, serverName);
                try {
                    if (opType.equals(DEPLOY)) {
                        doDeploy(connection,
                                targetOn,
                                opName,
                                opParams,
                                opSignature,
                                fileName);
                    } else if (opType.equals(UNDEPLOY)) {
                        doUnDeploy(connection,
                                targetOn,
                                opName,
                                opParams,
                                opSignature,
                                fileName);
                    } else if (opType.equals(UPLOAD_DEPLOY)) {
                        doUploadReDeploy(connection,
                                targetOn,
                                opName,
                                opParams,
                                opSignature,
                                fileName,
                                replaceExisting);
                    } else if (opType.equals(UPLOAD)) {
                        doUpload(connection,
                                targetOn,
                                fileName,
                                replaceExisting);
                    }
                } catch (Exception e) {
                    setProgress(FAILED);
                    HashMap status = new HashMap();
                    status.put(e.toString(), Boolean.valueOf(FAILURE));
                    progressReport.put(serverName, status);
                    continue;
                }
            } catch (Exception e) {
                setProgress(FAILED);
                HashMap status = new HashMap();
                status.put("Could not connect to: " +
                        serverName + ". Error: " + e.toString(), Boolean.valueOf(FAILURE));
                progressReport.put(serverName, status);
                continue;
            }
        }
        if (!getProgress().equals(FAILED)) {
            setProgress(COMPLETED);
        }
        return true;

    }

    /**
     * Deploy a module on a given server.
     *
     * @param connection MBeanServerConnection to use to invoke an MBean operation
     * @param targetOn ObjectName of the target server
     * @param opName the name of the deploy operation
     * @param opParams the parameters to this deploy operation
     * @param opSignature the method signature of the deploy operation
     * @param fileName the name of the module to deploy
     * @throws Exception if an exception is thrown during the undeployment
     */
    private void doDeploy(MBeanServerConnection connection, ObjectName targetOn,
            String opName, Object[] opParams, String[] opSignature,
            String fileName)
    throws Exception {

        String serverName = targetOn.getKeyProperty("name");

        if (!isDeployed(connection, targetOn, opName, opParams, opSignature)) {

            connection.invoke(targetOn, opName, opParams, opSignature);

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Opearation "
                        + opName + "'" + fileName
                        + ") on target " + serverName
                        + " done");
            }
            //progressReport.put(serverName, DEPLOYED);
            HashMap status = new HashMap();
            status.put(DEPLOYED, new Boolean(SUCCESS));
            progressReport.put(serverName, status);
        } else {
            //progressReport.put(serverName, "File is already deployed.");
            HashMap status = new HashMap();
            status.put("File is already deployed.", new Boolean(FAILURE));
            progressReport.put(serverName, status);
        }
    }

    /**
     * Undeploy a module on a given server.
     *
     * @param connection MBeanServerConnection to use to invoke an MBean operation
     * @param targetOn ObjectName of the target server
     * @param opName the name of the unDeploy operation
     * @param opParams the parameters to this undeploy operation
     * @param opSignature the method signature of the undeploy operation
     * @param fileName the name of the module to undeploy
     * @throws Exception if an exception is thrown during the deployment
     */
    private void doUnDeploy(MBeanServerConnection connection, ObjectName targetOn,
            String opName, Object[] opParams, String[] opSignature, String fileName)
    throws Exception {

        String serverName = targetOn.getKeyProperty("name");

        if (isDeployed(connection, targetOn, opName, opParams, opSignature)) {

            connection.invoke(targetOn, opName, opParams, opSignature);

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Opearation "
                        + opName + "'" + fileName
                        + ") on target " + serverName
                        + " done");
            }
            HashMap status = new HashMap();
            status.put(UNDEPLOYED, Boolean.valueOf(SUCCESS));
            progressReport.put(serverName, status);
        } else {
            HashMap status = new HashMap();
            status.put("File is not deployed.", Boolean.valueOf(FAILURE));
            progressReport.put(serverName, status);
        }
    }

    /**
     * Upload and (re)deploy a file on a target server.
     *
     * @param connection MBeanServerConnection to use to invoke an MBean operation
     * @param targetOn ObjectName of the target server
     * @param opName the name of the operation
     * @param opParams the parameters to this undeploy operation
     * @param opSignature the method signature of the undeploy operation
     * @param fileName the name of the module to upload and redeploy
     * @param replaceExisting If an existing file of the same name should be replaced.
     * @throws Exception if an exception is thrown during the deployment
     */
    private void doUploadReDeploy(MBeanServerConnection connection, ObjectName targetOn,
            String opName, Object[] opParams, String[] opSignature, String fileName, boolean replaceExisting)
    throws Exception {

        String serverName = targetOn.getKeyProperty("name");
        String uploadedName = fileName;

        try {
            uploadedName = doUpload(connection, targetOn, fileName, replaceExisting);
        } catch (Exception e) {
            HashMap status = new HashMap();
            status.put("File failed to upload: " + e.toString(), Boolean.valueOf(FAILURE));
            progressReport.put(serverName, status);
            return;
        }

        // the parameter needs to change to reflect the name of the module after
        // it was uploaded. So when we are uploading '/usr/share/java/some.jar'
        // it will need to be given to the deployJar command as 'deploy.jar'
        opParams[0] = uploadedName;

        boolean isDeployed = isDeployed(connection, targetOn, opName, opParams, opSignature);
        // redeploy an application because the module was replaced by a new one
        if (isDeployed && replaceExisting) {
            doUnDeploy(connection, targetOn, getUnDeployType(opName), opParams, opSignature, fileName);
            doDeploy(connection, targetOn, getDeployType(opName), opParams, opSignature, fileName);
        }
        // the uploaded module was not deployed, need to deploy it
        else if (!isDeployed) {
            doDeploy(connection,targetOn, getDeployType(opName), opParams, opSignature, fileName);
        } else {
            // Do nothing.
        }
    }

    /**
     * Uploads a file to a target server.
     *
     * @param connection MBeanServerConnection to use to invoke an MBean operation
     * @param targetOn ObjectName of the target server
     * @param fileName The file to upload
     * @param replaceExisting If an existing file of the same name should be replaced.
     * @return The path of the uploaded file.
     * @throws Exception if an exception is thrown during the deployment
     */
    private String doUpload(MBeanServerConnection connection, ObjectName targetOn, String fileName, boolean replaceExisting) throws Exception {

        String opName = "sendFile";

        File file = null;
        // first try the file name as it is
        file = new File(fileName).getCanonicalFile();
        String deployName = file.getName();
        if (!file.exists()) {
            // if the file was not found try looking for it under JONAS_BASE
            String dir = getFolderDir(fileName);
            file = new File(dir, fileName);
            deployName = fileName;
        }

        FileInputStream inputStream = new FileInputStream(file);
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        byte[] buf = new byte[BUFFER_SIZE];
        // Read bytes
        int len;
        while ((len = inputStream.read(buf)) > 0) {
            baos.write(buf, 0, len);
        }
        byte[] bytesOfFile = baos.toByteArray();

        Object[] opParams = new Object[]{bytesOfFile, deployName, Boolean.valueOf(replaceExisting)};
        String[] opSignature = new String[]{"[B", "java.lang.String", "boolean"};
        String filePath = (String) connection.invoke(targetOn, opName, opParams, opSignature);

        String serverName = targetOn.getKeyProperty("name");
        HashMap status = new HashMap();
        status.put("File uploaded to: '" + filePath + "'" , Boolean.valueOf(SUCCESS));
        progressReport.put(serverName, status);

        return deployName;
    }

    /**
     * Returns the directory that a file would be found in.
     *
     * @param fileName the file to find the directory for.
     * @return The directory of the given file.
     * @throws Exception if an exception is thrown during the deployment
     */
    private String getFolderDir(String fileName) throws Exception {
        String jBase = JProp.getJonasBase();
        // based on extension
        String dir = null;
        // EJB
        if (fileName.toLowerCase().endsWith(".jar")) {
            dir = jBase + File.separator + "ejbjars";
        } else if (fileName.toLowerCase().endsWith(".war")) {
            // War file
            dir = jBase + File.separator + "webapps";
        } else if (fileName.toLowerCase().endsWith(".ear")) {
            // ear file
            dir = jBase + File.separator + "apps";
        } else if (fileName.toLowerCase().endsWith(".rar")) {
            // rar file
            dir = jBase + File.separator + "rars";
        } else {
            // invalid type
            throw new Exception("Invalid extension for the file '" + fileName
                    + "'. Valid are .jar, .war, .ear, .rar");
        }
        return dir;
    }

    /**
     * If a file is already deployed on a target server
     *
     * @param connection the MBeanServerConnection to use to invoke the MBean operation
     * @param targetOn The target server name.
     * @param opName the name of the operation that was used to call this.
     * @param opParams the parameters to the MBean's invoke call.
     * @param opSignature the signature for the parameters.
     * @return if the given file is already deployed on the target server.
     * @throws Exception if an exception is thrown during the deployment
     */
    private boolean isDeployed(MBeanServerConnection connection,
            ObjectName targetOn, String opName, Object[] opParams,
            String[] opSignature) throws Exception {

        Boolean deployed = (Boolean) connection.invoke(targetOn,
                getIsDeployedType(opName), opParams, opSignature);

        return deployed.booleanValue();
    }

    /**
     * Returns which 'is deployed' method to use.
     * @param opName The operation being invoked.
     * @return Which 'is deployed' method to use.
     */
    private String getIsDeployedType(String opName) {
        String result = "";
        if (opName.indexOf("Ear") != -1) {
            result = "isEarDeployed";
        } else if (opName.indexOf("War") != -1) {
            result = "isWarDeployed";
        } else if (opName.indexOf("Rar") != -1) {
            result = "isRarDeployed";
        } else if (opName.indexOf("Jar") != -1) {
            result = "isJarDeployed";
        }
        return result;
    }

    /**
     * Depending on the name of the operation, which unDeploy method to use.
     * @param opName the operation name that was used to call this.
     * @return the right unDeploy method name to use.
     */
    private String getUnDeployType(String opName) {
        String result = "";
        if (opName.indexOf("Ear") != -1) {
            result = "unDeployEar";
        } else if (opName.indexOf("War") != -1) {
            result = "unDeployWar";
        } else if (opName.indexOf("Rar") != -1) {
            result = "unDeployRar";
        } else if (opName.indexOf("Jar") != -1) {
            result = "unDeployJar";
        }
        return result;
    }

    /**
     * Depending on the name of the operation, which deploy method to use.
     * @param opName the operation name that was used to call this.
     * @return the right deploy method name to use.
     */
    private String getDeployType(String opName) {
        String result = "";
        if (opName.indexOf("Ear") != -1) {
            result = "deployEar";
        } else if (opName.indexOf("War") != -1) {
            result = "deployWar";
        } else if (opName.indexOf("Rar") != -1) {
            result = "deployRar";
        } else if (opName.indexOf("Jar") != -1) {
            result = "deployJar";
        }
        return result;
    }

    /**
     * Check the current progress and determine whether a new deployment operation
     * is to be rejected or accepted.
     *
     * @return true if the operation should be rejected because another operation
     * is still running, false is the operation should be accepted and the
     * progress is set to RUNNING
     */
    private boolean isAnotherOperationRunning() {
        if (progress == null) {
            setProgress(RUNNING);
        } else {
            if (!getProgress().equals(RUNNING)) {
                setProgress(RUNNING);
            } else {
                return true;
            }
        }
        return false;
    }

    /**
     * Return true if all the elements in target are present in myServers list
     * @param target array of server OBJECT_NAMEs
     * @return true if all the elements in target are present in myServers list
     */
    private boolean isTargetCorrect(String[] target) {
        for (int i = 0; i < target.length; i++) {
            if (!myServers.contains(target[i])) {
                setGlobalErrorReport("Target element " + target[i] + " is not registered in J2EEDomain MBean " + getObjectName());
                return false;
            }
        }
        return true;
    }

    /**
     * Test a connection for server serverName
     * @param connection MBeanServerConnection to test
     * @param serverName the server's name
     * @throws IOException if thrown, the connection is faulty
     */
    private void testConnection(MBeanServerConnection connection, String serverName) throws Exception {
        if (connection != null) {
            Integer nb = connection.getMBeanCount();
        } else {
            // Maybe server stopped
            String state = getServerState(serverName);
            if (state != null) {
                throw new Exception("Target " + serverName + " state is " + state);
            } else {
                // The server was removed from managedServersToState table.
                // This happens if the server was added to the domain/cluster
                // by the Enroller and not by the DomainMap construction
                throw new Exception("Target " + serverName + " is no more under domain control");
            }
        }
    }

    private boolean isServerInServersList(String name, String[] serverNames) {
        boolean result = false;
        for (int i = 0; i < serverNames.length; i++) {
            if (serverNames[i].equals(name)) {
                return true;
            }
        }
        return result;
    }

    /**
     * @return Returns the jmxService.
     */
    public JmxService getJmxService() {
        return jmxService;
    }

    /**
     * @param jmxService The jmxService to set.
     */
    public void setJmxService(JmxService jmxService) {
        this.jmxService = jmxService;
    }

    /**
     * @return Returns the report.
     */
    public HashMap getProgressReport() {
        return progressReport;
    }

    /**
     * @param report The report to set.
     */
    public void setProgressReport(HashMap report) {
        this.progressReport = report;
    }

    /**
     * @return Returns the globalErrorReport.
     */
    public String getGlobalErrorReport() {
        return globalErrorReport;
    }

    /**
     * @param globalErrorReport The globalErrorReport to set.
     */
    public void setGlobalErrorReport(String globalErrorReport) {
        this.globalErrorReport = globalErrorReport;
    }

    /**
     * @return Returns the isCluster.
     */
    public boolean isCluster() {
        return isCluster;
    }

    /**
     * @return Returns the registeredOnsForListner.
     */
    private ArrayList getRegisteredOnsForListner() {
        return registeredOnsForListner;
    }

    /**
     * @param registeredOnsForListner The registeredOnsForListner to set.
     */
    private void setRegisteredOnsForListner(ArrayList registeredOnsForListner) {
        this.registeredOnsForListner = registeredOnsForListner;
    }

    /**
     * @return Returns the description.
     */
    public String getDescription() {
        return description;
    }

    /**
     * @param description The description to set.
     */
    public void setDescription(String description) {
        this.description = description;
    }

    /**
     * @return Returns the master.
     */
    public boolean isMaster() {
        return master;
    }

    /**
     * @param master The master to set.
     */
    public void setMaster(boolean master) {
        this.master = master;
    }
}
