/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2005 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.webapp.jonasadmin.deploy;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.objectweb.jonas.jmx.J2eeObjectName;
import org.objectweb.jonas.jmx.JonasManagementRepr;
import org.objectweb.jonas.webapp.jonasadmin.JonasAdminJmx;

/**
 * @author Patrick Smith
 * @author Gregory Lapouchnian
 * @author Adriana Danes
 */
public class ApplyDomainDeployConfirmAction extends BaseDeployAction {

    // --------------------------------------------------------- Properties

    private ArrayList targets;

    // --------------------------------------------------------- Public Methods

    public ActionForward executeAction(ActionMapping p_Mapping,
            ActionForm p_Form, HttpServletRequest p_Request,
            HttpServletResponse p_Response) throws IOException,
            ServletException {

        DomainDeployForm oForm = (DomainDeployForm) p_Form;

        // Create a new deployment thread only if there is no other deployment
        // in progress and not after a deployment has been completed.

        // These flags are in place to prevent multiple threads from being created
        // when this action is called again and again by the domainDeployProgress
        // JSP page.
        if (!oForm.getDeploymentInProgress() && !oForm.getDeploymentComplete()) {

            oForm.setDeploymentInProgress(true);

            // Create a new thread to deploy the applications to the servers
            // so that the progress JSP can be refreshed with updated progress.
            new DeployThread(p_Mapping, p_Form, p_Request, p_Response) {

                public void run() {
                    // Form used
                    DomainDeployForm oForm = (DomainDeployForm) p_Form;

                    try {
                        // apps to be deployed on these servers
                        ArrayList alApps = oForm.getListDeploy();
                        // deploy each app the set of selected servers
                        Map reportList = oForm.getReports();

                        String selectedAction = oForm.getSelectedAction();
                        boolean replaceExisting = oForm.getReplacementOption();

                        // selected targets (server and cluster ONs)
                        ArrayList targets = oForm.getListTargetsSelected();
                        // Map that gives for each selected domain/cluster the server targets to do deployment:
                        // - key : domain/cluster ON
                        // - value : ArrayList of server ONs.
                        HashMap clustServerMap = new HashMap();
                        Map updatedMap = getClusterServerMap(clustServerMap, targets, reportList, alApps);

                        for (int i = 0; i < alApps.size(); i++) {
                            // current application to deploy
                            String current = (String) alApps.get(i);
                            doSelectedOperation(current, selectedAction, replaceExisting, updatedMap, reportList);
                        }
                        oForm.setReports(reportList);

                        // there is no longer a deployment in progress
                        oForm.setDeploymentInProgress(false);
                        // this current deployment has been completed
                        oForm.setDeploymentComplete(true);
                    } catch (Throwable t) {
                        oForm.setException(true);
                        addGlobalError(t);
                        saveErrors(p_Request, m_Errors);
                    }
                }
            }.start();
        }

        // Forward to the jsp.
        return (p_Mapping.findForward("Domain Deploy Progress"));
    }

    private void createException(String message) throws OperationInProgressException {
        throw new OperationInProgressException(message);
    }

    /**
     * Do selected deployment action with the application named 'appName' on the selected targets
     * @param appName the name of the application
     * @param action action to do
     * @param replaceExisting true if must replace the file in case of a 'distribute' operation
     * @param clustServerMap Map that gives for each selected domain/cluster the server targets to do deployment
     * @param reportList to be updated by the deployment actions
     * @return true if the operation was executed, false if the operation was rejected
     */
    private boolean doSelectedOperation(String appName, String action, boolean replaceExisting, Map clustServerMap, Map reportList)
    throws OperationInProgressException {
        if (clustServerMap == null) {
            createException("Deployment operation rejected. Could not identified domain");
            return false;
        }
        boolean result = true;

        Iterator keys = clustServerMap.keySet().iterator();
        while (keys.hasNext()) {

            boolean clusterResult = true;
            ObjectName clustOn = (ObjectName) keys.next();
            ArrayList servers = (ArrayList) clustServerMap.get(clustOn);
            // attempt to deploy the application, save the report
            // to display to the user on the progress page and
            // reset the progress to allow the next deployment
            // to go through
            if (action.equals(DomainDeployForm.DEPLOY)) {
                clusterResult = doDeployOperation(appName, clustOn, servers);
            } else if (action.equals(DomainDeployForm.UPLOAD)) {
                clusterResult = doUploadOperation(appName, replaceExisting, clustOn, servers);
            } else if (action.equals(DomainDeployForm.UPLOADDEPLOY)) {
                clusterResult = doUploadDeployOperation(appName, replaceExisting, clustOn, servers);
            }

            if (clusterResult) {
                Map clustReport = getReport(clustOn);
                Map report = (Map) reportList.get(appName);
                if (report == null) {
                    reportList.put(appName, clustReport);
                } else {
                    report.putAll(clustReport);
                    reportList.put(appName, report);
                }
                resetProgress(clustOn);
            } else {
                createException("Deployment operation rejected on domain " + clustOn.toString());
            }

            result = result && clusterResult;
        }
        return result;
    }


    /**
     * Deploy the application with name 'appName' on the servers in the given domain/cluster
     * @param appName the name of the application
     * @param on given domain or cluster ObjectName
     * @param servers target servers in the domain or cluster
     * @return true if the operation was executed, false if the operation was
     * rejected
     */
    private boolean doDeployOperation(String appName, ObjectName on, ArrayList servers) {
        // store the servers in a String[]
        String[] serverArray = null;
        if (servers != null) {
            serverArray = new String[servers.size()];
            for (int i = 0; i < servers.size(); i++) {
                serverArray[i] = (String) servers.get(i);
            }
        }

        // define the paramters and the signature for the invocation
        String opName = getDomainDeploymentMethodName();
        Object[] params = new Object[] {serverArray, appName};
        String[] sig = new String[] {"[Ljava.lang.String;", "java.lang.String"};

        return ((Boolean) JonasManagementRepr.invoke(on, opName, params, sig)).booleanValue();
    }
    /**
     * Upload the application with name 'appName' on the servers in the given domain/cluster
     * @param appName the name of the application
     * @param replaceExisting if the upload operation should overwrite an existing file.
     * @param on given domain or cluster ObjectName
     * @param servers target servers in the domain or cluster
     * @return true if the operation was executed, false if the operation was
     * rejected
     */
    private boolean doUploadOperation(String appName, boolean replaceExisting, ObjectName on, ArrayList servers) {
        // store the servers in a String[]
        String[] serverArray = null;
        if (servers != null) {
            serverArray = new String[servers.size()];
            for (int i = 0; i < servers.size(); i++) {
                serverArray[i] = (String) servers.get(i);
            }
        }

        // define the paramters and the signature for the invocation
        String opName = "uploadFile";
        Object[] params = new Object[] {serverArray, appName, Boolean.valueOf(replaceExisting)};
        String[] sig = new String[] {"[Ljava.lang.String;", "java.lang.String", "boolean"};

        return ((Boolean) JonasManagementRepr.invoke(on, opName, params, sig)).booleanValue();
    }
    /**
     * Upload and redeploy the application with name 'appName' on the servers in the given domain/cluster
     * @param appName the application to upload and redeploy
     * @param replaceExisting if the upload operation should overwrite an existing file.
     * @param on given domain or cluster ObjectName
     * @param servers target servers in the domain or cluster
     * @return true if the operation was executed, false if the operation was
     * rejected
     */
    private boolean doUploadDeployOperation(String appName, boolean replaceExisting, ObjectName on, ArrayList servers) {
        // store the servers in a String[]
        String[] serverArray = null;
        if (servers != null) {
            serverArray = new String[servers.size()];
            for (int i = 0; i < servers.size(); i++) {
                serverArray[i] = (String) servers.get(i);
            }
        }

        // define the paramters and the signature for the invocation
        String opName = getDomainUploadDeployMethodName();
        Object[] params = new Object[] {serverArray, appName, Boolean.valueOf(replaceExisting)};
        String[] sig = new String[] {"[Ljava.lang.String;", "java.lang.String", "boolean"};

        return ((Boolean) JonasManagementRepr.invoke(on, opName, params, sig)).booleanValue();
    }
    /**
     * Invoke the resetProgress method of the J2EEDomain.
     * @param J2EEDoamin MBean ObjectName
     */
    private void resetProgress(ObjectName on) {
        Object[] params = new Object[] {};
        String[] sig = new String[] {};

        JonasManagementRepr.invoke(on, "resetProgress", params, sig);
    }

    /**
     * Get the report for the deployment operation from J2EEDomain MBean
     * @param J2EEDoamin MBean ObjectName
     * @return a HashMap containing the results of the deployment of an application
     * on each of the selected servers.
     */
    private HashMap getReport(ObjectName on) {
        HashMap report = (HashMap) JonasManagementRepr.getAttribute(on, "progressReport");

        return report;
    }

    /**
     * A custom thread class used for doing the deployment operations.
     */
    protected class DeployThread extends Thread {

        protected ActionMapping p_Mapping;

        protected ActionForm p_Form;

        protected HttpServletRequest p_Request;

        protected HttpServletResponse p_Response;

        public DeployThread(ActionMapping p_Mapping, ActionForm p_Form,
                HttpServletRequest p_Request, HttpServletResponse p_Response) {
            this.p_Mapping = p_Mapping;
            this.p_Form = p_Form;
            this.p_Request = p_Request;
            this.p_Response = p_Response;
        }

    }
    /**
     * Construct the clustServerMap, 1st phase: create data structures used by domain/cluster
     * hierarchy to update the clustServerMap.
     * @param clustServerMap
     * @parap targets list of target ONs
     * @param report report map to update
     * @param list of applications
     * @return a Map associating all the impacted domain/clusters to the target servers
     */
    private Map getClusterServerMap(HashMap clustServerMap, ArrayList targets, Map report, ArrayList alApps) {
        // walk down through the domain hierarchy to update the clustServerMap
        ObjectName domainOn = JonasAdminJmx.getJ2eeDomainObjectName();
        if (domainOn == null) {
            return null;
        }
        // List of domain OBJECT_NAMEs in selected targets
        ArrayList containersON = new ArrayList();
        // List of server OBJECT_NAMEs in selected targets
        ArrayList serversON = new ArrayList();
        // List of server names in selected targets
        ArrayList serverNames = new ArrayList();
        // For each target, determine the domain it belongs to
        for (int i = 0; i < targets.size(); i++) {
            String targetON = (String) targets.get(i);
            ObjectName targetOn = null;
            try {
                targetOn = ObjectName.getInstance(targetON);
            } catch (MalformedObjectNameException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
            if (targetOn == null) {
                continue;
            }
            if ("J2EEServer".equals(targetOn.getKeyProperty("j2eeType"))) {
                // the target is a server
                serversON.add(targetOn.toString());
                serverNames.add(targetOn.getKeyProperty("name"));
            } else if ("J2EEDomain".equals(targetOn.getKeyProperty("j2eeType"))) {
                // the target is a cluster
                containersON.add(targetOn.toString());
            }
        }
        prepareClusterServerMap(domainOn, clustServerMap, serversON, serverNames, containersON);
        return updateClusterMap(clustServerMap, report, alApps);
    }
    /**
     * Update the clustServerMap by mapping to the current domain/cluster its managed servers found
     * in he serverNames list. Its sub-clusters found un the containersON list are mapped to null, which
     * means that its entire contained is a taget. Also, the method is recurcevely called for each sub-cluster.
     * If the current domain/cluster its already contained in the map with null value, return. The mapping
     * will be updated by the updateClusterServerMap method.
     * @param on the current domain/cluster
     * @param clustServerMap the map to be updated
     * @param serversOn ONs of the selected servers
     * @param serverNames the name of the selected servers
     * @param containersON ONs of the selected clusters
     */
    private void prepareClusterServerMap(ObjectName on, HashMap clustServerMap, ArrayList serversOn, ArrayList serverNames, ArrayList containersON) {
        if (clustServerMap.keySet().contains(on.toString())) {
            addEmbeddedClusters(on, clustServerMap);
            return;
        }
        // get servers in the current domain/cluster
        String[] myNames = (String[]) JonasManagementRepr.getAttribute(on, "serverNames");
        // check for servers to deploy on
        ArrayList myServersToDeploy = new ArrayList();
        for (int i = 0; i < myNames.length; i++) {
            int index = serverNames.indexOf(myNames[i]);
            if (index >= 0) {
                // myNames[i] is the name of a target server to be deployed in the current domain/cluster
                myServersToDeploy.add(serversOn.get(index));
            }
        }
        if (!myServersToDeploy.isEmpty()) {
            clustServerMap.put(on.toString(), myServersToDeploy);
        }

        // get clusters in the current domain/cluster
        String[] myClusters = (String[]) JonasManagementRepr.getAttribute(on, "clusters");
        // check for clusters to deploy on
        for (int i = 0; i < myClusters.length; i++) {
            String clustON = myClusters[i];
            if (containersON.contains(clustON)) {
                // myClusters[i] is the ON of a target cluster
                clustServerMap.put(clustON, null); // the list of servers in this cluster is not set
            }
            // recursively call toto on myClusers
            ObjectName clustOn = null;
            try {
                clustOn = ObjectName.getInstance(clustON);
            } catch (MalformedObjectNameException e) {
                // TODO add debug trace
                // normally should never occur
            }
            if (clustOn != null) {
                prepareClusterServerMap(clustOn, clustServerMap, serversOn, serverNames, containersON);
            }
        }
    }
    /**
     * For a given cluster having associated null value in clustServerMap, recursivelly add
     * its sub-clusters with associated null values
     * @param on
     * @param clustServerMap
     */
    private void addEmbeddedClusters(ObjectName on, Map clustServerMap) {
        String[] myClusters = (String[]) JonasManagementRepr.getAttribute(on, "clusters");
        for (int i = 0; i < myClusters.length; i++) {
            String clustON = myClusters[i];
            clustServerMap.put(clustON, null);
            ObjectName clustOn = null;
            try {
                clustOn = ObjectName.getInstance(clustON);
            } catch (MalformedObjectNameException e) {
                // TODO add debug trace
                // normally should never occur
            }
            if (clustOn != null) {
                addEmbeddedClusters(clustOn, clustServerMap);
            }
        }
    }
    /**
     * Create a new clustServerMap having ObjectNames instead of ONs as keys and updated
     * values with the constraint that a server ON has to be present only once.
     * The null values are replaced by the list of all the managed server ONs, excepted
     * already registered ONs.
     * Once target server ONs collected, update the report Map
     * @param clustServerMap first version of map having null values and possibly multiple server ONs
     * @return a new Map having ObjectName keys and non null, unique values in the associated servers list.
     */
    private HashMap updateClusterMap(Map clustServerMap, Map report, ArrayList alApps) {
        HashMap resultMap = new HashMap();
        ArrayList alGlobalServersList = new ArrayList();

        Iterator keys = clustServerMap.keySet().iterator();
        while (keys.hasNext()) {
            String clustON = (String) keys.next();
            ObjectName on = null;
            try {
                on = ObjectName.getInstance(clustON);
            } catch (MalformedObjectNameException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
            if (on == null) {
                continue;
            }
            ArrayList servers = (ArrayList) clustServerMap.get(clustON);
            if (servers != null) {
                // check that servrs are not already in the alGlobalServersList
                for (int i = 0; i < servers.size(); i++) {
                    String server = (String) servers.get(i);
                    if (alGlobalServersList.contains(server)) {
                        servers.remove(i);
                    } else {
                        alGlobalServersList.add(server);
                    }
                }
            } else {
                servers = new ArrayList();
                String domainName = on.getDomain();
                String[] serverNames = (String[]) JonasManagementRepr.invoke(on, "getStartedServers", null, null);
                for (int i = 0; i < serverNames.length; i++) {
                    String serverName = serverNames[i];
                    String serverON = J2eeObjectName.J2EEServer(domainName, serverName).toString();
                    if (!alGlobalServersList.contains(serverON)) {
                        servers.add(serverON);
                        alGlobalServersList.add(serverON);
                    }
                }
            }
            resultMap.put(on, servers);
        }
        // update report
        Map initialStatus = new HashMap();
        initialStatus.put("In Progress...", "true");
        for (int i = 0; i < alApps.size(); i++) {
            HashMap appReport = new HashMap();
            for (int j = 0; j < alGlobalServersList.size(); j++) {
                ObjectName serverOn = null;
                try {
                    serverOn = ObjectName.getInstance((String) alGlobalServersList.get(j));
                } catch (MalformedObjectNameException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
                if (serverOn != null) {
                    appReport.put(serverOn.getKeyProperty("name"), initialStatus);
                }
            }
            report.put(alApps.get(i), appReport);
        }
        return resultMap;
    }
}
