/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 * Copyright (C) 2006 Distributed Systems Lab.
 * Universidad Polit??cnica de Madrid (Spain)
 * Contact: http://lsd.ls.fi.upm.es/lsd
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 *
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.io.ByteArrayOutputStream;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Arrays;

import javax.ejb.EJBException;
import javax.ejb.SessionBean;

import org.objectweb.carol.cmi.configuration.TraceCmi;
import org.objectweb.carol.cmi.ha.ReplicationException;
import org.objectweb.carol.cmi.ha.ReplicationManager;
import org.objectweb.carol.cmi.ha.RequestId;
import org.objectweb.carol.cmi.ha.ResponseInfo;


import org.objectweb.jonas.ha.HaUtilsDelegateImpl;
import org.objectweb.jonas.ha.interceptor.HACurrentDelegateImpl;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Generic part of the EJBObject implementation for replicated SFSBs
 * @author Francisco Perez-Sorrosal (fpsorrosal@no-spam@fi.upm.es)
 * @author Alberto Paz-Jimenez (apaz@no-spam@fi.upm.es)
 */
public abstract class JRepStatefulRemote extends JSessionRemote {

    private static ReplicationManager repMgr = null;

    public JRepStatefulRemote(JSessionFactory bf) throws RemoteException {
        super(bf);
        TraceCmi.debugCmiHA("JRepStatefulRemote");
        // Get a reference to the replication manager from the HA service
        if (repMgr == null) repMgr = HaUtilsDelegateImpl.getReplicationManager();
    }

    /**
     * preInvoke is called before any request.
     * @param txa Transaction Attribute (Supports, Required, ...)
     * @return A RequestCtx object
     * @throws RemoteException Thrown when the method failed due to a
     *         system-level failure.
     */
    public RequestCtx preInvoke(int txa) throws RemoteException {
        if (TraceEjb.isDebugIc()) {
            TraceEjb.interp.log(BasicLevel.DEBUG, "");
        }
        RequestCtx rctx = bf.preInvokeRemote(txa);
        bs.setMustCommit(rctx.mustCommit); // for remove stateful session
        bs.enlistConnections(rctx.currTx); // Enlist connection list to tx
        /* ******************REPLICATION CODE STARTS HERE*********************** */
        preInvokeHook(rctx); // Replication algorithm link
        /* **********************END OF REPLICATION CODE************************ */
        return rctx;
    }

    /**
     * postInvoke is called after any request.
     * @param rctx The RequestCtx that was returned at preInvoke()
     * @param remove TODO
     * @parem response The response that is going to be returned to the client
     *        (Needed by HA)
     * @throws RemoteException Thrown when the method failed due to a
     *         system-level failure.
     */
    public void postInvoke(RequestCtx rctx, Object response, boolean remove) throws RemoteException {
        if (TraceEjb.isDebugIc()) {
            TraceEjb.interp.log(BasicLevel.DEBUG, "");
        }
        /* ******************REPLICATION CODE STARTS HERE*********************** */
        try {
            postInvokeHook(rctx, response, remove); // Replication algorithm
                                                    // link
        /* **********************END OF REPLICATION CODE************************ */
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            bs.delistConnections(rctx.currTx);
            // save current tx (for Bean Managed only)
            bs.saveBeanTx();
            try {
                bf.postInvokeRemote(rctx);
            } finally {
                bs.releaseICtx(rctx, rctx.sysExc != null);
            }
        }
    }

    /* ******************REPLICATION CODE STARTS HERE*********************** */

    // ------------------------------------------------------------------------
    // Protected methods ------------------------------------------------------
    // ------------------------------------------------------------------------
    /**
     * Injects the state of the bean if the cluster object id received in the
     * request is in the current node. It extracts the object id passed through
     * the interceptors from the request id included in the current HA context.
     * @return true if the bean state has been injected
     */
    protected ResponseInfo injectState() {
        ResponseInfo response = null;
        HACurrentDelegateImpl current = HACurrentDelegateImpl.getCurrent();
        RequestId reqId = current.getCurrentReq();
        if (reqId != null) {
            response = repMgr.restoreBeanChanges(reqId, (JStatefulSwitch) this.bs);
        }
        return response;
    }

    // ------------------------------------------------------------------------
    // Private methods --------------------------------------------------------
    // ------------------------------------------------------------------------

    /**
     * preInvokeHook implements the hook to perform the required replication
     * tasks before the bean context has been established. In this case, it
     * saves the current state of the bean before perform the invocation, in the
     * request context.
     * @param rctx The RequestCtx that was returned at preInvoke()
     */
    private void preInvokeHook(RequestCtx rctx) {
        try {
            JStatefulSwitch sfsw = ((JStatefulSwitch) bs);
            rctx.state = getStatefulState(sfsw);
        } catch (EJBException e) {
            e.printStackTrace();
        }
    }

    /**
     * postInvokeHook implements the hook to perform the required replication
     * tasks after the bean context has been established
     * @param rctx The RequestCtx that was returned at preInvoke()
     * @parem response The response that is going to be returned to the client
     * @param remove true if it is an invocation to a remove method
     */
    private void postInvokeHook(RequestCtx rctx, Object response, boolean remove) {
        TraceCmi.debugCmiHA("---------------------------------------");
        TraceCmi.debugCmiHA("This is the PostInvoke hook...");
        if (remove) {
            replicateRemove();
        } else {
            replicateRequest(rctx, response);
        }
        TraceCmi.debugCmiHA("---------------------------------------");
    }

    /**
     * Called if we have to replicate a remove() invocation
     */
    private void replicateRemove() {
        HACurrentDelegateImpl current = HACurrentDelegateImpl.getCurrent();
        try {
            repMgr.replicateRemove(current.getCurrentReq());
        } catch (ReplicationException e) {
            TraceCmi.error("Cannot replicate removal for: " + current.getCurrentReq());
        }
    }

    /**
     * Called if we have to replicate a request
     */
    private void replicateRequest(RequestCtx rctx, Object response) {
        // Create the rootId from the current HACtx/Thread association
        HACurrentDelegateImpl current = HACurrentDelegateImpl.getCurrent();
        RequestId rootId = current.getRootReq();
        // Get the previous and the current state
        JStatefulSwitch sfsw = ((JStatefulSwitch) bs);
        byte[] previousState = rctx.state;
        byte[] state = getStatefulState(sfsw);
        // If the state has changed, hold the new bean state on HA service
        if (!Arrays.equals(previousState, state)) {
            if (TraceCmi.isDebugCmiHA()) {
                TraceCmi.debugCmiHA("The state HAS changed... Adding bean to req changes...");
                TraceCmi.debugCmiHA("\tBean: " + current.getCurrentReq().getObjectId());
            }
            repMgr.addModifiedBean(rootId, current.getCurrentReq().getObjectId(), state);
        }
        try {
            // Check if we have to replicate
            if ((current.replicate() == true)) {
                repMgr.replicate(rootId, response);
            } else {
                TraceCmi.debugCmiHA("Nested call... Delaying replication");
            }
        } catch (ReplicationException e) {
            TraceCmi.error("Cannot replicate request: " + rootId);
        }
    }

    /**
     * Extracts the state of the stateful session bean
     * @param jss the related switch of the stateful session bean to access the
     *        SFBS's real instance
     */
    private byte[] getStatefulState(JStatefulSwitch jss) {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        JStatefulContext ctx = jss.getStatefulContext();
        try {
            SessionBean instance = ctx.getInstance();
            JStatefulOutputStream out = new JStatefulOutputStream(bos);
            out.writeObject(instance);
            out.close();
        } catch (Exception e) {
            TraceCmi.error("Can't get state of " + bf.getEJBName() + "inst.");
        }
        return bos.toByteArray();
    }

    /* **********************END OF REPLICATION CODE************************ */
}
